// =============================================================================
//
//   glUnrealMesh
//   Unit for loading Unreal1-Meshes (vertex animations)
//
// =============================================================================
//   Copyright  2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the >>GNU Public License (GPL)<< which
//   can be obtained here : http://opensource.org/licenses/gpl-license.php
//   So only use this file if you fully unterstand that license!!!
// =============================================================================
//   Version history :
//    Release 0.1 - Initial Release, it may contain some bugs since the
//                  unreal mesh format is almost undocumented
// =============================================================================
//   I took most of the infos on the file format from here :
//      http://www.unrealized.com/main/articles.html?id=924475569
//   and here :
//      http://astronomy.swin.edu.au/~pbourke/geomformats/unreal/
// =============================================================================
//   NOTE : This WON'T WORK WITH LODMESHES FROM UNREAL1 V225. Only normal
//          meshes can be loaded. LodMeshes are not documented at all.
// =============================================================================

unit glUnrealMesh;

interface

uses
 Windows,
 Classes,
 SysUtils,
 Textures,
 dglOpenGL;


type
 TVertex = record
   x,y,z : Single;
  end;

 TUnreal_DataHeader = packed record
   NumPolygons : Word;
   NumVertices : Word;
   BogusRot    : Word;
   BogusFrame  : Word;
   BogusNormX  : DWord;
   BogusNormY  : DWord;
   BogusNormZ  : DWord;
   FixScale    : DWord;
   Unused      : array[0..2] of DWord;
   Unknown     : array[0..11] of Char;
  end;
 TUnreal_Triangle = packed record
   VertexIndex : array[0..2] of Word;
   mType       : Byte;
   mColor      : Byte;
   TexCoords   : array[0..2, 0..1] of Byte;
   mTextureNum : Byte;
   mFlags      : Byte;
  end;
 TUnreal_FrameData = record
   VertexPos    : array of TVertex;
   VertexNormal : array of TVertex;
  end;
 TUnreal_Sequence = record
   StartFrame : Word;
   NumFrames  : Word;
   Rate       : Single;
  end;

 TUnrealMesh = class(TComponent)
  private
   procedure CalculateVertexNormals;
  public
   Triangle     : array of TUnreal_Triangle;
   Frame        : array of TUnreal_FrameData;
   Sequence     : array of TUnreal_Sequence;
   Normal       : array of TVertex;
   SequenceName : TStringList;
   DataHeader   : TUnreal_DataHeader;
   NumFrames    : Word;
   FrameSize    : Word;
   CurrFrame    : Single;
   CurrSequence : Word;
   Texture      : glUInt;
   procedure Render;
   procedure ProgressSequence(pTimeFactor : Single);
   procedure LoadFromFile(pDataFile, pAnimFile, pUnrealScript, pSkinTexture : String);
   constructor Create(AOwner : TComponent); reintroduce;
   destructor Destroy; override;
  end;

implementation

function Vertex(pX, pY, pZ : Single) : TVertex;
begin
Result.x := pX;
Result.y := pY;
Result.z := pZ;
end;

// Converts a string to a float but replaces the '.' with the decimal seperator
function StrToFloat(pString : String) : Single;
var
 i : Integer;
begin
for i := 1 to Length(pString) do
 if pString[i] = '.' then
  pString[i] := DecimalSeparator;
Result := SysUtils.StrToFloat(pString);
end;

function ExtractStr(pSourceString, pStartToken : String; pEndChar : Char) : String;
var
 i : Integer;
begin
Result := '';
if Pos(pStartToken, pSourceString) = 0 then
 exit;
i := Pos(pStartToken, pSourceString)+Length(pStartToken);
repeat
Result := Result+pSourceString[i];
inc(i);
if i > Length(pSourceString) then
 break;
until pSourceString[i] = pEndChar;
if Result = '' then
 raise Exception.Create('ExtractStr->Could not extract string!');
end;

// =============================================================================
//  TUnrealMesh
// =============================================================================

// =============================================================================
//  TUnrealMesh.Create                                           - CONSTRUCTOR -
// =============================================================================
constructor TUnrealMesh.Create(AOwner : TComponent);
begin
inherited;
SequenceName := TStringList.Create;
end;

// =============================================================================
//  TUnrealMesh.Destroy                                           - DESTRUCTOR -
// =============================================================================
destructor TUnrealMesh.Destroy;
begin
SequenceName.Free;
inherited;
end;

// =============================================================================
//  TUnrealMesh.Render
// =============================================================================
procedure TUnrealMesh.Render;
var
 t : Integer;
begin
if DataHeader.NumPolygons = 0 then
 exit;
CalculateVertexNormals;
glBegin(GL_TRIANGLES);
 for t := 0 to DataHeader.NumPolygons-2 do
  if Triangle[t].mType <= 4 then
   begin
   glTexCoord2f(Triangle[t].TexCoords[0,0]/255, Triangle[t].TexCoords[0,1]/255);
   glNormal3f(Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[0]].x,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[0]].y,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[0]].z);
   glVertex3f(Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[0]].x,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[0]].y,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[0]].z);

   glTexCoord2f(Triangle[t].TexCoords[1,0]/255, Triangle[t].TexCoords[1,1]/255);
   glNormal3f(Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[1]].x,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[1]].y,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[1]].z);
   glVertex3f(Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[1]].x,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[1]].y,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[1]].z);

   glTexCoord2f(Triangle[t].TexCoords[2,0]/255, Triangle[t].TexCoords[2,1]/255);
   glNormal3f(Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[2]].x,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[2]].y,
              Frame[Round(CurrFrame)].VertexNormal[Triangle[t].VertexIndex[2]].z);
   glVertex3f(Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[2]].x,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[2]].y,
              Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[2]].z);
   end;
glEnd;
end;

// =============================================================================
//  TUnrealMesh.ProgressSequence
// =============================================================================
procedure TUnrealMesh.ProgressSequence(pTimeFactor : Single);
begin
if DataHeader.NumPolygons = 0 then
 exit;
CurrFrame := CurrFrame + pTimeFactor * Sequence[CurrSequence].Rate / 30;
if CurrFrame > Sequence[CurrSequence].StartFrame+Sequence[CurrSequence].NumFrames-1 then
 CurrFrame := Sequence[CurrSequence].StartFrame;
end;

// =============================================================================
//  TUnrealMesh.LoadFromFile
// =============================================================================
procedure TUnrealMesh.LoadFromFile(pDataFile, pAnimFile, pUnrealScript, pSkinTexture : String);
var
 PackedVertex : Integer;
 F            : TFileStream;
 S            : TStringList;
 TempStr      : String;
 i,v          : Integer;
begin
// ==== Data file ==============================================================
// Header
F := TFileStream.Create(pDataFile, fmOpenRead);
F.Read(DataHeader, SizeOf(DataHeader));
// Triangle data
SetLength(Triangle, DataHeader.NumPolygons);
for i := 0 to DataHeader.NumPolygons-1 do
 begin
 F.Read(Triangle[i], SizeOf(TUnreal_Triangle));
 end;
F.Free;
// ==== Aniv file ==============================================================
F := TFileStream.Create(pAnimFile, fmOpenRead);
// Number of frames
F.Read(NumFrames, SizeOf(Word));
// Framesize
F.Read(FrameSize, SizeOf(Word));
// Framedata
SetLength(Frame, NumFrames);
for i := 0 to NumFrames-1 do
 begin
 SetLength(Frame[i].VertexPos, DataHeader.NumVertices);
 for v := 0 to DataHeader.NumVertices-1 do
  begin
  F.Read(PackedVertex, SizeOf(Integer));
  // Unpack Vertrex
  Frame[i].VertexPos[v].x := ( (PackedVertex and $7FF) div 8);
  Frame[i].VertexPos[v].y := ( ((PackedVertex shr 11) and $7FF) div 8);
  Frame[i].VertexPos[v].z := ( ((PackedVertex shr 22) and $3FF) div 4);
  if Frame[i].VertexPos[v].y > 128 then
   Frame[i].VertexPos[v].y := Frame[i].VertexPos[v].y - 256;
  if Frame[i].VertexPos[v].x  > 128 then
   Frame[i].VertexPos[v].x := Frame[i].VertexPos[v].x - 256;
  if Frame[i].VertexPos[v].z > 128 then
   Frame[i].VertexPos[v].z := Frame[i].VertexPos[v].z - 256;
  end;
 end;
//
F.Free;
// ==== UnrealScript ===========================================================
if FileExists(pUnrealScript) then
 begin
 S := TStringList.Create;
 S.LoadFromFile(pUnrealScript);
 SequenceName.Clear;
 CurrFrame    := 0;
 CurrSequence := 0;
 SetLength(Sequence, 0);
 // Extract sequence names, startframe and duration
 for i := 0 to S.Count-1 do
  if Pos('MESH SEQUENCE', S[i]) > 0 then
   begin
   SetLength(Sequence, Length(Sequence)+1);
   SequenceName.Add(ExtractStr(S[i], 'SEQ=', ' '));
   Sequence[High(Sequence)].StartFrame := StrToInt(ExtractStr(S[i], 'STARTFRAME=', ' '));
   Sequence[High(Sequence)].NumFrames  := StrToInt(ExtractStr(S[i], 'NUMFRAMES=', ' '));
   TempStr := ExtractStr(S[i], 'RATE=', ' ');
   if TempStr > '' then
    Sequence[High(Sequence)].Rate := StrToFloat(TempStr)
   else
    Sequence[High(Sequence)].Rate := 30; // Default rate for animations seems to be 30 FpS
   end;
 S.Free;
 end
else
 begin
 // No unrealscript -> insert just one animationsequence that contains all frames
 SequenceName.Add('all');
 SetLength(Sequence, 1);
 Sequence[0].StartFrame := 0;
 Sequence[0].NumFrames  := NumFrames;
 end;
// ==== Skintexture ============================================================
LoadTexture(pSkinTexture, Texture);
end;

// =============================================================================
//  TUnrealMesh.CalculateVertexNormals
// =============================================================================
procedure TUnrealMesh.CalculateVertexNormals;
var
 TriNormals : array of TVertex;
 t,v,i      : Integer;
 v1,v2,v3,n : TVertex;
begin
SetLength(TriNormals, DataHeader.NumPolygons);
// Calculate per-Triangle-Normals first
for t  := 0 to High(TriNormals) do
 begin
 v1  := Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[0]];
 v2  := Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[1]];
 v3  := Frame[Round(CurrFrame)].VertexPos[Triangle[t].VertexIndex[2]];
 v2  := Vertex(-v2.x, -v2.y, -v2.z);
 v1  := Vertex(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z);
 v3  := Vertex(v3.x+v2.x, v3.y+v2.y, v3.z+v2.z);
 n.x := (v1.y*v3.z) - (v1.z*v3.y);
 n.y := (v1.z*v3.x) - (v1.x*v3.z);
 n.z := (v1.x*v3.y) - (v1.y*v3.x);
 n.x := n.x / Sqrt(n.x*n.x + n.y*n.y + n.z*n.z);
 n.y := n.y / Sqrt(n.x*n.x + n.y*n.y + n.z*n.z);
 n.z := n.z / Sqrt(n.x*n.x + n.y*n.y + n.z*n.z);
 TriNormals[t] := n;
 end;
// Vertexnormal
SetLength(Frame[Round(CurrFrame)].VertexNormal, DataHeader.NumVertices);
for v := 0 to High(Frame[Round(CurrFrame)].VertexNormal) do
 begin
 Frame[Round(CurrFrame)].VertexNormal[v] := Vertex(0,0,0);
 for t := 0 to High(TriNormals) do
  for i := 0 to 2 do
   if Triangle[t].VertexIndex[i] = v then
    begin
    Frame[Round(CurrFrame)].VertexNormal[v].x := Frame[Round(CurrFrame)].VertexNormal[v].x + TriNormals[t].x;
    Frame[Round(CurrFrame)].VertexNormal[v].y := Frame[Round(CurrFrame)].VertexNormal[v].y + TriNormals[t].y;
    Frame[Round(CurrFrame)].VertexNormal[v].z := Frame[Round(CurrFrame)].VertexNormal[v].z + TriNormals[t].z;
    end;
 with Frame[Round(CurrFrame)].VertexNormal[v] do
  begin
  x := x / Sqrt(x*x + y*y + z*z);
  y := y / Sqrt(x*x + y*y + z*z);
  z := z / Sqrt(x*x + y*y + z*z);
  end;
 end;
end;


end.
